import numpy as np
from dataclasses import dataclass
from acados_template import AcadosModel
from casadi import MX, vertcat, exp, fmax


@dataclass
class CSTRParameters:
    F0: float = 0.1  # m^3/min
    T0: float = 350.0  # K
    c0: float = 1.0  # kmol/m^3
    r: float = 0.219  # m
    k0: float = 7.2 * 1e10  # 1/min
    EbR: float = 8750  # K
    U: float = 54.94  # kJ / (min*m^2*K)
    rho: float = 1000  # kg / m^3
    Cp: float = 0.239  # kJ / (kg*K)
    dH: float = -5 * 1e4  # kJ / kmol
    xs: np.ndarray = np.array([0.878, 324.5, 0.659])
    us: np.ndarray = np.array([300, 0.1])


def setup_cstr_model_equations(params: CSTRParameters):

    # set up states
    c = MX.sym("c")  # molar concentration of species A
    T = MX.sym("T")  # reactor temperature
    h = MX.sym("h")  # level of the tank

    x = vertcat(c, T, h)

    # controls
    Tc = MX.sym("Tc")  # temperature of coolant liquid
    F = MX.sym("F")    # outlet flowrate

    u = vertcat(Tc, F)

    # xdot
    c_dot = MX.sym("c_dot")
    T_dot = MX.sym("T_dot")
    h_dot = MX.sym("h_dot")

    xdot = vertcat(c_dot, T_dot, h_dot)

    # dynamics
    A_const = np.pi * params.r**2
    denom = A_const * (fmax(h, 1e-5))
    k = params.k0 * exp(-params.EbR / T)
    rate = k * c

    f_expl = vertcat(
        params.F0 * (params.c0 - c) / denom - rate,
        params.F0 * (params.T0 - T) / denom
        - params.dH / (params.rho * params.Cp) * rate
        + 2 * params.U / (params.r * params.rho * params.Cp) * (Tc - T),
        (params.F0 - F) / A_const,
    )

    return x, u, xdot, f_expl


def setup_cstr_model(params: CSTRParameters):

    model_name = "cstr_ode"

    x, u, xdot, f_expl = setup_cstr_model_equations(params)

    f_impl = xdot - f_expl

    model = AcadosModel()

    model.f_impl_expr = f_impl
    model.f_expl_expr = f_expl
    model.x = x
    model.xdot = xdot
    model.u = u
    model.name = model_name

    return model
